#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#if HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#include <errno.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#if HAVE_INTTYPES_H
#include <inttypes.h>
#endif
#if HAVE_LIBGEN_H
#include <libgen.h>
#endif

#include "lion.h"
#include "misc.h"

#define DEBUG_FLAG 4096
#include "debug.h"
#include "httpd.h"
#include "parser.h"
#include "request.h"
#include "upnp.h"
#include "conf.h"

#if WIN32
#include "win32.h"
#endif





void upnp_setcommand(request_t *node, char *cmd)
{

    if (!cmd || !*cmd) return;

    debugf("[upnp] set command '%s'\n", cmd);

    if (!mystrccmp("llink.xml", cmd)) {
        // Change the request to the download of the skin/upnp/llink.xml file
        SAFE_FREE(node->path);
        node->path = strdup("/llink.xml");
        debugf("[upnp] changed to llink.xml get\n");
    }

}

void upnp_parseinput(request_t *node, char *line)
{
	char *r, *path;

    debugf("[upnp] input '%s'\n", line);

    // FIXME, this needs to do real parsing. For now, we
    // do some ugly things and guess what they want. Yay.

    if (strstr(line, "u:GetSortCapabilities")) {
        debugf("[upnp] guessing GetSortCapabilities!\n");
        node->upnp_type = UPNP_GETSORTCAPABILITIES;
    }
    if (strstr(line, "BrowseDirectChildren")) {
        debugf("[upnp] guessing BrowseDirectChildren!\n");
        node->upnp_type = UPNP_BROWSEDIRECTCHILDREN;
    }

	// What directory do they want, pray tell?
	//<ObjectID>00Lokien</ObjectID>
    // xbox bug sends "ContainerID" ?
    if ((r = strstr(line, "<ObjectID>")) ||
        (r = strstr(line, "<ContainerID>"))) {

		// Skip past obj-id
        r = strchr(r, '>');
        if (!r) return;
        r++;
		// get the path, ignoring leading whitespace, up to "<" or EOL
		// which ever is first. Spaces in paths will be URL encoded. Yay!
		//path = misc_digtoken(&r, " <\r\n");
		path = misc_digtoken(&r, " <&?\r\n");
		if (path && *path) {
			char *ar, *token;
			// UPnPs root is 0.
			if (!mystrccmp("0", path)) *path = '/';

			debugf("[upnp] Directory wanted '%s'\n", path);
			SAFE_FREE(node->path);

			node->path = request_url_decode(path);
			// We need to check for "CGI" part in the URL here too.

			if ((misc_digtoken_optchar == '&') ||
				(misc_digtoken_optchar == '?')) {

				token = misc_digtoken(&r, " <\r\n");
				// Check if it has a "&". If so, that's cgi stuff.
				if (token && *token) {

					SAFE_COPY(node->cgi, token);

				}
			}

			if (node->cgi)
				request_cgi_parse(node);

			debugf("[upnp] Directory: '%s' cgi: '%s'\n",
                   node->path,
                   node->cgi ? node->cgi : "");


		}
	} // ObjectID


	//<StartingIndex>0</StartingIndex>' (remaining 152 - read 32)
    if ((r = strstr(line, "<StartingIndex>"))) {
        debugf("[upnp] guessing StartingIndex!\n");
		node->page_from = atoi(&r[15]);
    }


	//<RequestedCount>20</RequestedCount>' (remaining 118 - read
    if ((r = strstr(line, "<RequestedCount>"))) {
        debugf("[upnp] guessing RequestedCount!\n");
        // This needs to be page_from + RequestedCount, but we can't rely on
        // the order in the XML, so we add them up later.
		node->page_to = atoi(&r[16]);
    }


}


void upnp_reply(request_t *node)
{
	int sent = 0;

    debugf("[upnp] reply\n");

    // Sanity, should not happen
    if (!node->tmphandle) return;

    switch(node->upnp_type) {
    case UPNP_GETSORTCAPABILITIES:
		debugf("[upnp] GetSortCapabilities reply\n");
#if 0
        lion_printf(node->tmphandle,
                    "HTTP/1.1 200 OK\r\n"
                    "Content-Length: 374\r\n"
                    "CONTENT-TYPE: text/xml; charset=\"utf-8\"\r\n"
                    "Server: llink-%s (I'm just making stuff up)\r\n"
                    "\r\n"
                    VERSION);
#endif

        lion_printf(node->tmphandle,
                    "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n"
                    "<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\"><s:Body>\r\n"
                    "<u:GetSortCapabilitiesResponse xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\">\r\n"
                    "<SortCaps>dc:title, upnp:originalTrackNumber</SortCaps>\r\n"
                    "</u:GetSortCapabilitiesResponse>\r\n"
                    "</s:Body> </s:Envelope>\r\n\r\n"
                    );

		debugf(
			   "200 OK\r\n"
			   "Content-Length: 374\r\n"
			   "CONTENT-TYPE: text/xml; charset=\"utf-8\"\r\n"
			   "Server: llink-%s (I'm just making stuff up)\r\n"
			   "\r\n"
			   "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n"
			   "<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\"><s:Body>\r\n"
			   "<u:GetSortCapabilitiesResponse xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\">\r\n"
			   "<SortCaps>dc:title,                         </SortCaps>\r\n"
			   "</u:GetSortCapabilitiesResponse>\r\n"
			   "</s:Body> </s:Envelope>\r\n\r\n",
			   VERSION);

        break;

	case UPNP_BROWSEDIRECTCHILDREN:
		debugf("[upnp] Browse: \n");



#if 0
		lion_printf(node->tmphandle,
					"HTTP/1.1 200 OK\r\n"
					"Content-Type: text/xml; charset=\"utf-8\"\r\n"
					"Content-Length: 1036\r\n"
					"Server: llink-%s (Browse)\r\n"
					"\r\n",
					VERSION);

		sent = lion_printf(node->tmphandle,
					"<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n"
					"<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\"><s:Body><u:BrowseResponse xmlns:u=\"urn:schemas-upnp-org:service:ContentDirectory:1\"><Result>&lt;DIDL-Lite xmlns:dc=\"http://purl.org/dc/elements/1.1/\" xmlns:upnp=\"urn:schemas-upnp-org:metadata-1-0/upnp/\" xmlns=\"urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/\"&gt;\r\n"
					"&lt;item id=\"64$4$0$0\" parentID=\"64$4$0\" restricted=\"1\"&gt;&lt;dc:title&gt;MyFilename&lt;/dc:title&gt;&lt;upnp:class&gt;object.item.videoItem&lt;/upnp:class&gt;&lt;dc:creator&gt;DiVX&lt;/dc:creator&gt;&lt;dc:date&gt;2009-06-28T03:06:21&lt;/dc:date&gt;&lt;res size=\"733581312\" duration=\"1:32:20.076\" bitrate=\"132413\" resolution=\"624x352\" protocolInfo=\"http-get:*:video/divx:*\"&gt;http://192.168.0.254:8200/MediaItems/51.avi&lt;/res&gt;&lt;/item&gt;&lt;/DIDL-Lite&gt;</Result>\r\n"
					"<NumberReturned>1</NumberReturned>\r\n"
					"<TotalMatches>1</TotalMatches>\r\n"
					"<UpdateID>2</UpdateID></u:BrowseResponse></s:Body></s:Envelope>\r\n"
					"\r\n"
					);
		debugf("[upnp] sent %d.\n", sent);
#endif
        break;

    default:
        break;

    }

    // Unknown, error, who cares?
#if 0
    lion_printf(node->handle,
                "HTTP/1.1 200 OK\r\n"
                "Server: llink-%s\r\n"
                "Content-Length: 0\r\n"
                "\r\n",
				VERSION);
#endif


    debugf("[upnp] finished. Handing back to request\n");
    request_finishdir(node);

}


